# Copyright 2020 Codethink Ltd
# Distributed under the terms of the Modified BSD License.

import abc
import base64
import calendar
import contextlib
import datetime
import hashlib
import logging
import os
import os.path as path
import shutil
import tempfile
import warnings
from contextlib import contextmanager
from os import PathLike
from threading import Lock
from typing import IO, List, Tuple, Union

import aiofiles
import aioshutil
import fsspec
from tenacity import retry
from tenacity.retry import retry_if_exception_type
from tenacity.stop import stop_after_attempt

try:
    import xattr

    has_xattr = True
except ImportError:
    has_xattr = False
    warnings.warn("xattr not installed, not using it for file metadata")

from quetz.errors import ConfigError

File = IO[bytes]

StrPath = Union[str, PathLike]

logger = logging.getLogger("quetz")


class PackageStore(abc.ABC):
    def __init__(self):
        self._download_locks = {}

    def get_download_lock(self, channel: str, destination: str):
        return self._download_locks.get((channel, destination))

    def create_download_lock(self, channel: str, destination: str):
        lock = self._download_locks[(channel, destination)] = Lock()
        return lock

    def delete_download_lock(self, channel: str, destination: str):
        del self._download_locks[(channel, destination)]

    @property
    def kind(self):
        return type(self).__name__

    @property
    @abc.abstractmethod
    def support_redirect(self) -> bool:
        pass

    @abc.abstractmethod
    def create_channel(self, name):
        pass

    @abc.abstractmethod
    def remove_channel(self, name):
        """remove channel recursively"""
        pass

    @abc.abstractmethod
    def list_files(self, channel: str) -> List[str]:
        pass

    @abc.abstractmethod
    def url(self, channel: str, src: str, expires: int = 0) -> str:
        pass

    @abc.abstractmethod
    def add_package_async(self, package: File, channel: str, destination: str):
        pass

    @abc.abstractmethod
    def add_package(self, package: File, channel: str, destination: str):
        pass

    @abc.abstractmethod
    def add_file(
        self, data: Union[str, bytes], channel: str, destination: StrPath
    ) -> None:
        pass

    @abc.abstractmethod
    def serve_path(self, channel, src):
        pass

    @abc.abstractmethod
    def delete_file(self, channel: str, destination: str):
        """remove file from package store"""

    @abc.abstractmethod
    def move_file(self, channel: str, source: str, destination: str):
        """move file from source to destination in package store"""

    @abc.abstractmethod
    def file_exists(self, channel: str, destination: str):
        """Return True if the file exists"""

    @abc.abstractmethod
    def get_filemetadata(self, channel: str, src: str) -> Tuple[int, int, str]:
        """get file metadata: returns (file size, last modified time, etag)"""

    @abc.abstractclassmethod
    def cleanup_temp_files(self, channel: str, dry_run: bool = False):
        """clean up temporary `*.json{HASH}.[bz2|gz]` files from pkgstore"""


# generate a secret token for use with nginx secure link
# similar to https://stackoverflow.com/a/52764346 (thanks @flix on stackoverflow)
def nginx_secure_link(url: str, secret: str, expires=3600):
    future = datetime.datetime.utcnow() + datetime.timedelta(seconds=expires)
    expire_ts = calendar.timegm(future.timetuple())

    hash_string = f"{expire_ts}{url} {secret}".encode('utf-8')
    m = hashlib.md5(hash_string).digest()

    base64_hash = base64.urlsafe_b64encode(m)
    base64_s = base64_hash.decode('utf-8').rstrip('=')
    return base64_s, expire_ts


class LocalStore(PackageStore):
    def __init__(self, config):
        self.fs: fsspec.AbstractFileSystem = fsspec.filesystem("file")
        self.channels_dir = config['channels_dir']
        self.redirect_enabled = config['redirect_enabled']
        self.redirect_endpoint = config['redirect_endpoint']
        self.redirect_secret = config.get('redirect_secret')
        self.redirect_expiration = config.get('redirect_expiration')

        super().__init__()

    @property
    def support_redirect(self):
        return self.redirect_enabled

    @contextmanager
    def _atomic_open(self, channel: str, destination: StrPath, mode="wb") -> IO:
        full_path = path.join(self.channels_dir, channel, destination)
        self.fs.makedirs(path.dirname(full_path), exist_ok=True)

        # Creates a tempfile in the same directory as the target filename.
        # Renames it into place when it's closed.
        fh, tmpname = tempfile.mkstemp(dir=os.path.dirname(full_path), prefix=".")
        f = open(fh, mode)
        try:
            yield f
        except:  # noqa
            f.close()
            os.remove(tmpname)
            raise
        else:
            f.flush()  # Belt and braces (network file systems)
            f.close()
            os.rename(tmpname, full_path)

    def create_channel(self, name):
        self.fs.makedirs(path.join(self.channels_dir, name), exist_ok=True)

    def remove_channel(self, name):
        channel_path = path.join(self.channels_dir, name)
        self.fs.rm(channel_path, recursive=True)

    def add_package(self, package: File, channel: str, destination: str) -> None:
        with self._atomic_open(channel, destination) as f:
            shutil.copyfileobj(package, f)

    async def add_package_async(
        self, package: File, channel: str, destination: str
    ) -> None:
        full_path = path.join(self.channels_dir, channel, destination)
        self.fs.makedirs(path.dirname(full_path), exist_ok=True)

        async with aiofiles.open(full_path, 'wb') as f:
            await f.write(package.read())

    def add_file(
        self, data: Union[str, bytes], channel: str, destination: StrPath
    ) -> None:
        mode = "w" if isinstance(data, str) else "wb"
        with self._atomic_open(channel, destination, mode) as f:
            f.write(data)

    def delete_file(self, channel: str, destination: str):
        self.fs.delete(path.join(self.channels_dir, channel, destination))

    def move_file(self, channel: str, source: str, destination: str):
        self.fs.move(
            path.join(self.channels_dir, channel, source),
            path.join(self.channels_dir, channel, destination),
        )

    def file_exists(self, channel: str, destination: str):
        return self.fs.exists(path.join(self.channels_dir, channel, destination))

    def serve_path(self, channel, src):
        return self.fs.open(path.join(self.channels_dir, channel, src))

    def list_files(self, channel: str):
        channel_dir = os.path.join(self.channels_dir, channel)
        return [os.path.relpath(f, channel_dir) for f in self.fs.find(channel_dir)]

    def url(self, channel: str, src: str, expires=0):
        if self.redirect_enabled:
            # generate url + secret if necessary
            if self.redirect_secret:
                url = path.join(self.channels_dir, channel, src)
                md5hash, expires_by = nginx_secure_link(
                    url, self.redirect_secret, self.redirect_expiration
                )
                return (
                    path.join(self.redirect_endpoint, url)
                    + f"?md5={md5hash}&expires={expires_by}"
                )

            return path.join(self.redirect_endpoint, self.channels_dir, channel, src)
        else:
            return path.abspath(path.join(self.channels_dir, channel, src))

    def get_filemetadata(self, channel: str, src: str):
        filepath = path.abspath(path.join(self.channels_dir, channel, src))
        if not path.exists(filepath):
            raise FileNotFoundError()

        stat_res = os.stat(filepath)
        mtime = stat_res.st_mtime
        msize = stat_res.st_size

        xattr_failed = False
        etag = ""
        if has_xattr:
            try:
                # xattr will fail here if executing on e.g. the tmp filesystem
                attrs = xattr.xattr(filepath)
                attrs['user.testifpermissionok'] = b''
                try:
                    etag = attrs['user.etag'].decode('ascii')
                except KeyError:
                    # calculate md5 sum
                    with self.fs.open(filepath, 'rb') as f:
                        etag = hashlib.md5(f.read()).hexdigest()
                    attrs['user.etag'] = etag.encode('ascii')
            except OSError:
                xattr_failed = True

        elif not has_xattr or xattr_failed:
            etag_base = str(mtime) + "-" + str(msize)
            etag = hashlib.md5(etag_base.encode()).hexdigest()
        else:
            raise RuntimeError

        return (msize, mtime, etag)

    def cleanup_temp_files(self, channel: str, dry_run: bool = False):
        temp_files = []
        for each_end in [".bz2", ".gz"]:
            temp_files.extend(
                self.fs.glob(
                    f"{path.join(self.channels_dir, channel)}/**/*.json?*{each_end}"
                )
            )

        for each_temp_file in temp_files:
            logger.info(f"removing {each_temp_file} from pkgstore")
            if not dry_run:
                self.fs.delete(each_temp_file)


class S3Store(PackageStore):
    def __init__(self, config):
        try:
            import s3fs
        except ModuleNotFoundError:
            raise ModuleNotFoundError("S3 package store requires s3fs module")

        client_kwargs = {}
        url = config.get('url')
        region = config.get("region")
        if url:
            client_kwargs['endpoint_url'] = url
        if region:
            client_kwargs["region_name"] = region

        # When using IAM, key and secret will be empty, so need to pass None
        # to the s3fs constructor
        key = config['key'] if config['key'] != '' else None
        secret = config['secret'] if config['secret'] != '' else None
        self._key = key
        self._secret = secret
        self._client_kwargs = client_kwargs
        self.fs = s3fs.S3FileSystem(
            key=key,
            secret=secret,
            client_kwargs=client_kwargs,
        )

        self.bucket_prefix = config['bucket_prefix']
        self.bucket_suffix = config['bucket_suffix']
        super().__init__()

    @property
    def support_redirect(self):
        return True

    @contextlib.contextmanager
    def _get_fs(self):
        try:
            yield self.fs
        except PermissionError as e:
            raise ConfigError(f"{e} - check configured S3 credentials")

    def _bucket_map(self, name):
        return f"{self.bucket_prefix}{name}{self.bucket_suffix}"

    def create_channel(self, name):
        """Create the bucket if one doesn't already exist

        Parameters
        ----------
        name : str
            The name of the bucket to create on s3
        """
        with self._get_fs() as fs:
            try:
                fs.mkdir(self._bucket_map(name), acl="private")
            except FileExistsError:
                pass

    # we need to retry due to eventual (vs strong) consistency on Openstack Swift
    @retry(stop=stop_after_attempt(5), retry=retry_if_exception_type(OSError))
    def remove_channel(self, name):
        channel_path = self._bucket_map(name)
        self.fs.rm(channel_path, recursive=True, acl="private")

    def add_package(self, package: File, channel: str, destination: str) -> None:
        with self._get_fs() as fs:
            bucket = self._bucket_map(channel)
            with fs.open(path.join(bucket, destination), "wb", acl="private") as pkg:
                # use a chunk size of 10 Megabytes
                shutil.copyfileobj(package, pkg, 10 * 1024 * 1024)

    async def add_package_async(
        self, package: File, channel: str, destination: str
    ) -> None:
        import s3fs

        fs = s3fs.S3FileSystem(
            key=self._key,
            secret=self._secret,
            client_kwargs=self._client_kwargs,
        )
        bucket = self._bucket_map(channel)
        with fs.open(path.join(bucket, destination), "wb", acl="private") as pkg:
            # use a chunk size of 10 Megabytes
            await aioshutil.copyfileobj(package, pkg, 10 * 1024 * 1024)

    def add_file(
        self, data: Union[str, bytes], channel: str, destination: StrPath
    ) -> None:
        if type(data) is str:
            mode = "w"
        else:
            mode = "wb"

        with self._get_fs() as fs:
            bucket = self._bucket_map(channel)
            with fs.open(path.join(bucket, destination), mode, acl="private") as f:
                f.write(data)

    def serve_path(self, channel, src):
        with self._get_fs() as fs:
            return fs.open(path.join(self._bucket_map(channel), src))

    def delete_file(self, channel: str, destination: str):
        channel_bucket = self._bucket_map(channel)

        with self._get_fs() as fs:
            fs.delete(path.join(channel_bucket, destination))

    def move_file(self, channel: str, source: str, destination: str):
        channel_bucket = self._bucket_map(channel)

        with self._get_fs() as fs:
            fs.move(
                path.join(channel_bucket, source),
                path.join(channel_bucket, destination),
            )

    def file_exists(self, channel: str, destination: str):
        channel_bucket = self._bucket_map(channel)
        with self._get_fs() as fs:
            return fs.exists(path.join(channel_bucket, destination))

    def list_files(self, channel: str):
        def remove_prefix(text, prefix):
            if text.startswith(prefix):
                return text[len(prefix) :].lstrip("/")  # noqa: E203
            return text

        channel_bucket = self._bucket_map(channel)

        with self._get_fs() as fs:
            return [remove_prefix(f, channel_bucket) for f in fs.find(channel_bucket)]

    def url(self, channel: str, src: str, expires=3600):
        # expires is in seconds, so the default is 60 minutes!
        with self._get_fs() as fs:
            return fs.url(path.join(self._bucket_map(channel), src), expires)

    def get_filemetadata(self, channel: str, src: str):
        with self._get_fs() as fs:
            filepath = path.join(self._bucket_map(channel), src)
            infodata = fs.info(filepath)

            mtime = infodata['LastModified'].timestamp()
            msize = infodata['size']
            etag = infodata['ETag']

            return (msize, mtime, etag)

    def cleanup_temp_files(self, channel: str, dry_run: bool = False):
        with self._get_fs() as fs:
            temp_files = []
            for each_end in [".bz2", ".gz"]:
                temp_files.extend(
                    fs.glob(f"{self._bucket_map(channel)}/**/*.json?*{each_end}")
                )

            for each_temp_file in temp_files:
                logger.info(f"removing {each_temp_file} from pkgstore")
                if not dry_run:
                    fs.delete(each_temp_file)


class AzureBlobStore(PackageStore):
    def __init__(self, config):
        try:
            import adlfs
        except ModuleNotFoundError:
            raise ModuleNotFoundError("Azure Blob package store requires adlfs module")

        self.storage_account_name = config.get('account_name')
        self.access_key = config.get("account_access_key")
        self.conn_string = config.get("conn_str")

        self.fs = adlfs.AzureBlobFileSystem(
            account_name=self.storage_account_name,
            connection_string=self.conn_string,
            account_key=self.access_key,
        )

        self.container_prefix = config['container_prefix']
        self.container_suffix = config['container_suffix']
        super().__init__()

    @property
    def support_redirect(self):
        return True

    @contextlib.contextmanager
    def _get_fs(self):
        try:
            yield self.fs
        except PermissionError as e:
            raise ConfigError(f"{e} - check configured Azure Blob credentials")

    def _container_map(self, name):
        return f"{self.container_prefix}{name}{self.container_suffix}"

    def create_channel(self, name):
        """Create the container if one doesn't already exist

        Parameters
        ----------
        name : str
            The name of the container to create on azure blob
        """
        with self._get_fs() as fs:
            try:
                fs.mkdir(self._container_map(name))
            except FileExistsError:
                pass

    def remove_channel(self, name):
        channel_path = self._container_map(name)
        with self._get_fs() as fs:
            fs.rm(channel_path, recursive=True)

    def add_package(self, package: File, channel: str, destination: str) -> None:
        with self._get_fs() as fs:
            container = self._container_map(channel)
            with fs.open(path.join(container, destination), "wb") as pkg:
                # use a chunk size of 10 Megabytes
                shutil.copyfileobj(package, pkg, 10 * 1024 * 1024)

    async def add_package_async(
        self, package: File, channel: str, destination: str
    ) -> None:
        import adlfs

        fs = adlfs.AzureBlobFileSystem(
            account_name=self.storage_account_name,
            connection_string=self.conn_string,
            account_key=self.access_key,
        )
        container = self._container_map(channel)
        with fs.open(path.join(container, destination), "wb") as pkg:
            # use a chunk size of 10 Megabytes
            await aioshutil.copyfileobj(package, pkg, 10 * 1024 * 1024)

    def add_file(
        self, data: Union[str, bytes], channel: str, destination: StrPath
    ) -> None:
        if type(data) is str:
            mode = "w"
        else:
            mode = "wb"

        with self._get_fs() as fs:
            container = self._container_map(channel)
            with fs.open(path.join(container, destination), mode) as f:
                f.write(data)

    def serve_path(self, channel, src):
        with self._get_fs() as fs:
            return fs.open(path.join(self._container_map(channel), src))

    def delete_file(self, channel: str, destination: str):
        channel_container = self._container_map(channel)

        with self._get_fs() as fs:
            fs.delete(path.join(channel_container, destination))

    def move_file(self, channel: str, source: str, destination: str):
        channel_container = self._container_map(channel)

        with self._get_fs() as fs:
            fs.move(
                path.join(channel_container, source),
                path.join(channel_container, destination),
            )

    def file_exists(self, channel: str, destination: str):
        channel_container = self._container_map(channel)
        with self._get_fs() as fs:
            return fs.exists(path.join(channel_container, destination))

    def list_files(self, channel: str):
        def remove_prefix(text, prefix):
            if text.startswith(prefix):
                return text[len(prefix) :].lstrip("/")  # noqa: E203
            return text

        channel_container = self._container_map(channel)

        with self._get_fs() as fs:
            return [
                remove_prefix(f, channel_container) for f in fs.find(channel_container)
            ]

    def url(self, channel: str, src: str, expires=3600):
        # expires is in seconds, so the default is 60 minutes!
        with self._get_fs() as fs:
            return fs.url(path.join(self._container_map(channel), src), expires)

    def get_filemetadata(self, channel: str, src: str):
        with self._get_fs() as fs:
            filepath = path.join(self._container_map(channel), src)
            infodata = fs.info(filepath)

            mtime = infodata['last_modified'].timestamp()
            msize = infodata['size']
            etag = infodata['etag']

            return (msize, mtime, etag)

    def cleanup_temp_files(self, channel: str, dry_run: bool = False):
        with self._get_fs() as fs:
            temp_files = []
            for each_end in [".bz2", ".gz"]:
                temp_files.extend(
                    fs.glob(f"{self._container_map(channel)}/**/*.json?*{each_end}")
                )

            for each_temp_file in temp_files:
                logger.info(f"removing {each_temp_file} from pkgstore")
                if not dry_run:
                    fs.delete(each_temp_file)


class GoogleCloudStorageStore(PackageStore):
    def __init__(self, config):
        try:
            import gcsfs
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                "Google Cloud Storage package store requires gcsfs module"
            )

        self.project = config.get("project")
        self.token = config.get("token")
        self.cache_timeout = config.get("cache_timeout")
        self.region = config.get("region")

        self.fs = gcsfs.GCSFileSystem(
            project=self.project,
            token=self.token if self.token else None,
            cache_timeout=self.cache_timeout,
            default_location=self.region,
        )

        self.bucket_prefix = config['bucket_prefix']
        self.bucket_suffix = config['bucket_suffix']
        super().__init__()

    @property
    def support_redirect(self):
        # `gcsfs` currently doesnt support signing yet. Once this is implemented we
        # can enable this again.
        return True

    @contextlib.contextmanager
    def _get_fs(self):
        try:
            yield self.fs
        except PermissionError as e:
            raise ConfigError(
                f"{e} - check configured Google Cloud Storage credentials"
            )

    def _bucket_map(self, name):
        return f"{self.bucket_prefix}{name}{self.bucket_suffix}"

    def create_channel(self, name):
        """Create the container if one doesn't already exist

        Parameters
        ----------
        name : str
            The name of the container to create on azure blob
        """
        with self._get_fs() as fs:
            bucket_name = self._bucket_map(name)
            if f"{bucket_name}/" not in fs.buckets:
                fs.mkdir(bucket_name)
                fs.invalidate_cache()

    def remove_channel(self, name):
        channel_path = self._bucket_map(name)
        with self._get_fs() as fs:
            fs.rm(channel_path, recursive=True)

    def add_package(self, package: File, channel: str, destination: str) -> None:
        with self._get_fs() as fs:
            container = self._bucket_map(channel)
            with fs.open(path.join(container, destination), "wb") as pkg:
                # use a chunk size of 10 Megabytes
                shutil.copyfileobj(package, pkg, 10 * 1024 * 1024)

    async def add_package_async(
        self, package: File, channel: str, destination: str
    ) -> None:
        import gcsfs

        fs = gcsfs.GCSFileSystem(
            project=self.project,
            token=self.token if self.token else None,
            cache_timeout=self.cache_timeout,
            default_location=self.region,
        )
        container = self._bucket_map(channel)
        with fs.open(path.join(container, destination), "wb") as pkg:
            # use a chunk size of 10 Megabytes
            await aioshutil.copyfileobj(package, pkg, 10 * 1024 * 1024)

    def add_file(
        self, data: Union[str, bytes], channel: str, destination: StrPath
    ) -> None:
        if type(data) is str:
            mode = "w"
        else:
            mode = "wb"

        with self._get_fs() as fs:
            container = self._bucket_map(channel)
            with fs.open(path.join(container, destination), mode) as f:
                f.write(data)

    def serve_path(self, channel, src):
        with self._get_fs() as fs:
            file = fs.open(path.join(self._bucket_map(channel), src))
            info = file.info()
            if info["type"] != "file":
                raise FileNotFoundError()
            return file

    def delete_file(self, channel: str, destination: str):
        channel_container = self._bucket_map(channel)

        with self._get_fs() as fs:
            fs.delete(path.join(channel_container, destination))

    def move_file(self, channel: str, source: str, destination: str):
        channel_container = self._bucket_map(channel)

        with self._get_fs() as fs:
            fs.move(
                path.join(channel_container, source),
                path.join(channel_container, destination),
            )

    def file_exists(self, channel: str, destination: str):
        channel_container = self._bucket_map(channel)
        with self._get_fs() as fs:
            return fs.exists(path.join(channel_container, destination))

    def list_files(self, channel: str):
        def remove_prefix(text, prefix):
            if text.startswith(prefix):
                return text[len(prefix) :].lstrip("/")  # noqa: E203
            return text

        channel_container = self._bucket_map(channel)

        with self._get_fs() as fs:
            return [
                remove_prefix(f, channel_container) for f in fs.find(channel_container)
            ]

    def url(self, channel: str, src: str, expires=3600):
        # expires is in seconds, so the default is 60 minutes!
        with self._get_fs() as fs:
            expiration_timestamp = (
                int(datetime.datetime.now(tz=datetime.timezone.utc).timestamp())
                + expires
            )
            redirect_url = fs.sign(
                path.join(self._bucket_map(channel), src), expiration_timestamp
            )
            return redirect_url

    def get_filemetadata(self, channel: str, src: str):
        with self._get_fs() as fs:
            filepath = path.join(self._bucket_map(channel), src)
            infodata = fs.info(filepath)

            if infodata['type'] != 'file':
                raise FileNotFoundError()

            mtime = datetime.datetime.fromisoformat(
                infodata['updated'].replace('Z', '+00:00')
            ).timestamp()
            msize = infodata['size']
            etag = infodata['etag']

            return (msize, mtime, etag)

    def cleanup_temp_files(self, channel: str, dry_run: bool = False):
        with self._get_fs() as fs:
            temp_files = []
            for each_end in [".bz2", ".gz"]:
                temp_files.extend(
                    fs.glob(f"{self._bucket_map(channel)}/**/*.json?*{each_end}")
                )

            for each_temp_file in temp_files:
                logger.info(f"removing {each_temp_file} from pkgstore")
                if not dry_run:
                    fs.delete(each_temp_file)



def test_docs_endpoint(client):
    response = client.get("/docs")
    assert response.status_code == 200
    assert "text/html" in response.headers['Content-Type']


def test_openapi_endpoint(client):
    response = client.get("/openapi.json")
    assert response.status_code == 200
    assert "application/json" in response.headers['Content-Type']


def test_redoc_endpoint(client):
    response = client.get("/redoc")
    assert response.status_code == 200
    assert "text/html" in response.headers['Content-Type']



"""adding url and platforms dirs

Revision ID: 0653794b6252
Revises: 53f81aba78ce
Create Date: 2021-01-11 18:18:44.530101

"""
import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = '0653794b6252'
down_revision = '53f81aba78ce'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('aggregated_metrics', schema=None) as batch_op:
        batch_op.alter_column(
            'period',
            existing_type=postgresql.INTERVAL(),
            type_=sa.Enum('hour', 'day', 'month', 'year', name='intervaltype'),
            existing_nullable=True,
        )

    with op.batch_alter_table('packages', schema=None) as batch_op:
        batch_op.add_column(sa.Column('platforms', sa.String(), nullable=True))
        batch_op.add_column(sa.Column('url', sa.String(), nullable=True))

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('packages', schema=None) as batch_op:
        batch_op.drop_column('url')
        batch_op.drop_column('platforms')

    with op.batch_alter_table('aggregated_metrics', schema=None) as batch_op:
        batch_op.alter_column(
            'period',
            existing_type=sa.Enum('hour', 'day', 'month', 'year', name='intervaltype'),
            type_=postgresql.INTERVAL(),
            existing_nullable=True,
        )

    # ### end Alembic commands ###



import datetime
import uuid

import pytest
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm.exc import ObjectDeletedError

from quetz import errors, rest_models
from quetz.dao import Dao
from quetz.database import get_session
from quetz.db_models import Channel, Package, PackageVersion
from quetz.metrics.db_models import IntervalType, PackageVersionMetric, round_timestamp


@pytest.fixture
def channel(dao, db, user, channel_name):
    channel_data = rest_models.Channel(name=channel_name, private=False)
    channel = dao.create_channel(channel_data, user.id, "owner")
    yield channel

    try:
        db.delete(channel)
        db.commit()
    except ObjectDeletedError:
        pass


@pytest.fixture
def package(dao, channel, package_name, user, db):
    package_data = rest_models.Package(name=package_name)

    package = dao.create_package(channel.name, package_data, user.id, "owner")

    yield package

    db.delete(package)
    db.commit()


@pytest.fixture
def package_version(dao, package, user):
    return dao.create_version(
        channel_name=package.channel_name,
        package_name=package.name,
        package_format="tarbz2",
        platform="noarch",
        version="0.0.1",
        build_number="0",
        build_string="",
        filename="filename.tar.bz2",
        info="{}",
        uploader_id=user.id,
        size=101,
        upsert=False,
    )


def test_channel_with_huge_size_limit(dao, user, db):
    channel_data = rest_models.Channel(
        name="my-channel", private=False, size_limit=1000000000000000000
    )
    channel = dao.create_channel(channel_data, user.id, "owner", 1000000000000000000)
    del channel


def test_create_channel_with_invalid_name(dao, user, db):
    with pytest.raises(errors.ValidationError):
        channel_data = rest_models.Channel(name="my_channel", private=False)
        channel = dao.create_channel(channel_data, user.id, "owner")
        del channel


def test_create_version(dao, package, channel_name, package_name, db, user):
    assert (
        not db.query(PackageVersion)
        .filter(PackageVersion.package_name == package_name)
        .first()
    )
    assert dao.db == db
    dao.create_version(
        channel_name=channel_name,
        package_name=package_name,
        package_format="tarbz2",
        platform="noarch",
        version="0.0.1",
        build_number="0",
        build_string="",
        filename="filename.tar.bz2",
        info="{}",
        uploader_id=user.id,
        size=0,
        upsert=False,
    )

    created_version = (
        db.query(PackageVersion)
        .filter(PackageVersion.package_name == package_name)
        .first()
    )

    assert created_version
    assert created_version.version == "0.0.1"
    assert created_version.build_number == 0
    assert created_version.filename == "filename.tar.bz2"
    assert created_version.info == "{}"
    assert created_version.time_created == created_version.time_modified

    # error for insert-only with existing row
    with pytest.raises(IntegrityError):
        dao.create_version(
            channel_name=channel_name,
            package_name=package_name,
            package_format="tarbz2",
            platform="noarch",
            version="0.0.1",
            build_number="0",
            build_string="",
            filename="filename-2.tar.bz2",
            info="{}",
            size=0,
            uploader_id=user.id,
            upsert=False,
        )

    # update with upsert
    dao.create_version(
        channel_name=channel_name,
        package_name=package_name,
        package_format="tarbz2",
        platform="noarch",
        version="0.0.1",
        build_number="0",
        build_string="",
        filename="filename-2.tar.bz2",
        info='{"version": "x.y.z"}',
        uploader_id=user.id,
        size=0,
        upsert=True,
    )

    created_version = (
        db.query(PackageVersion)
        .filter(PackageVersion.package_name == package_name)
        .first()
    )

    assert created_version
    assert created_version.version == "0.0.1"
    assert created_version.build_number == 0
    assert created_version.filename == "filename-2.tar.bz2"
    assert created_version.info == '{"version": "x.y.z"}'
    assert created_version.time_created != created_version.time_modified


def test_update_channel(dao, channel, db):
    assert not channel.private
    dao.update_channel(channel.name, {"private": True})

    channel = db.query(Channel).filter(Channel.name == channel.name).one()

    assert channel.private


def test_update_channel_size(dao, channel, db, package_version):
    dao.update_channel_size(channel.name)

    channel = db.query(Channel).filter(Channel.name == channel.name).one()

    assert channel.size == package_version.size


def test_increment_download_count(
    dao: Dao, channel, db, package_version, session_maker
):
    assert package_version.download_count == 0
    now = datetime.datetime(2020, 10, 1, 10, 1, 10)
    dao.incr_download_count(
        channel.name, package_version.filename, package_version.platform, timestamp=now
    )

    download_counts = db.query(PackageVersionMetric).all()
    for m in download_counts:
        assert m.count == 1

    assert len(download_counts) == len(IntervalType)

    db.refresh(package_version)
    assert package_version.download_count == 1

    dao.incr_download_count(
        channel.name, package_version.filename, package_version.platform, timestamp=now
    )
    download_counts = db.query(PackageVersionMetric).all()
    for m in download_counts:
        assert m.count == 2

    assert len(download_counts) == len(IntervalType)

    db.refresh(package_version)
    assert package_version.download_count == 2

    dao.incr_download_count(
        channel.name,
        package_version.filename,
        package_version.platform,
        timestamp=now + datetime.timedelta(days=1),
    )

    download_counts = db.query(PackageVersionMetric).all()
    assert len(download_counts) == len(IntervalType) + 2

    db.refresh(package_version)
    assert package_version.download_count == 3


def test_get_package_version_metrics(dao: Dao, channel, db, package_version):
    now = datetime.datetime(2020, 10, 1, 10, 1, 10)
    dao.incr_download_count(
        channel.name, package_version.filename, package_version.platform, timestamp=now
    )

    metrics = dao.get_package_version_metrics(
        package_version.id, IntervalType.hour, "download"
    )

    metrics_dict = [(m.timestamp, m.count) for m in metrics]
    timestamp = now.replace(minute=0, second=0)

    assert metrics_dict == [(timestamp, 1)]

    hour = datetime.timedelta(hours=1)
    day = datetime.timedelta(days=1)

    metrics = dao.get_package_version_metrics(
        package_version.id,
        IntervalType.hour,
        "download",
        start=now - hour,
        end=now + hour,
    )

    metrics_dict = [(m.timestamp, m.count) for m in metrics]
    assert metrics_dict == [(timestamp, 1)]

    metrics = dao.get_package_version_metrics(
        package_version.id,
        IntervalType.hour,
        "download",
        start=now - hour,
        end=now + hour,
        fill_zeros=True,
    )

    metrics_dict = [(m.timestamp, m.count) for m in metrics]
    assert metrics_dict == [
        (timestamp - hour, 0),
        (timestamp, 1),
        (timestamp + hour, 0),
    ]

    # no start/end

    metrics = dao.get_package_version_metrics(
        package_version.id,
        IntervalType.hour,
        "download",
        start=now - hour,
        fill_zeros=True,
    )
    metrics_dict = [(m.timestamp, m.count) for m in metrics]
    assert metrics_dict == [
        (timestamp - hour, 0),
        (timestamp, 1),
    ]

    metrics = dao.get_package_version_metrics(
        package_version.id,
        IntervalType.hour,
        "download",
        end=now + hour,
        fill_zeros=True,
    )
    metrics_dict = [(m.timestamp, m.count) for m in metrics]
    assert metrics_dict == [
        (timestamp, 1),
        (timestamp + hour, 0),
    ]

    metrics = dao.get_package_version_metrics(
        package_version.id,
        IntervalType.hour,
        "download",
        fill_zeros=True,
    )
    metrics_dict = [(m.timestamp, m.count) for m in metrics]
    assert metrics_dict == [
        (timestamp, 1),
    ]

    # day interval
    timestamp_day = timestamp.replace(hour=0)
    metrics = dao.get_package_version_metrics(
        package_version.id, IntervalType.day, "download"
    )
    metrics_dict = [(m.timestamp, m.count) for m in metrics]
    assert metrics_dict == [(timestamp_day, 1)]

    metrics = dao.get_package_version_metrics(
        package_version.id,
        IntervalType.day,
        "download",
        start=now - day,
        end=now + day,
        fill_zeros=True,
    )

    metrics_dict = [(m.timestamp, m.count) for m in metrics]
    assert metrics_dict == [
        (timestamp_day - day, 0),
        (timestamp_day, 1),
        (timestamp_day + day, 0),
    ]

    # two items
    dao.incr_download_count(
        channel.name,
        package_version.filename,
        package_version.platform,
        timestamp=now + datetime.timedelta(hours=2),
    )

    metrics = dao.get_package_version_metrics(
        package_version.id,
        IntervalType.hour,
        "download",
        fill_zeros=True,
    )
    metrics_dict = [(m.timestamp, m.count) for m in metrics]
    assert metrics_dict == [
        (timestamp, 1),
        (timestamp + hour, 0),
        (timestamp + 2 * hour, 1),
    ]


@pytest.mark.parametrize("interval", list(IntervalType))
def test_get_package_version_metrics_intervals(
    dao: Dao, channel, db, package_version, interval
):
    now = datetime.datetime(2020, 10, 1, 10, 1, 10)
    dao.incr_download_count(
        channel.name, package_version.filename, package_version.platform, timestamp=now
    )

    metrics = dao.get_package_version_metrics(package_version.id, interval, "download")
    timestamp_interval = round_timestamp(now, interval)
    metrics_dict = [(m.timestamp, m.count) for m in metrics]
    assert metrics_dict == [(timestamp_interval, 1)]

    end = timestamp_interval.replace(year=2021)
    metrics = dao.get_package_version_metrics(
        package_version.id,
        interval,
        "download",
        start=timestamp_interval,
        end=end,
        fill_zeros=True,
    )

    metrics_dict = [(m.timestamp, m.count) for m in metrics]

    assert metrics_dict[0] == (timestamp_interval, 1)
    assert metrics_dict[-1] == (end, 0)


def test_create_user_with_profile(dao: Dao, user_without_profile):
    user = dao.create_user_with_profile(
        user_without_profile.username,
        provider="github",
        identity_id="1",
        name="new user",
        avatar_url="http://avatar",
        emails=None,
        role=None,
        exist_ok=True,
    )

    assert user.profile

    with pytest.raises(IntegrityError):
        dao.create_user_with_profile(
            user_without_profile.username,
            provider="github",
            identity_id="1",
            name="new user",
            avatar_url="http://avatar",
            role=None,
            exist_ok=False,
        )


@pytest.fixture
def db_extra(database_url):
    """a separate session for db connection

    Use only for tests that require two sessions concurrently.
    For most cases you will want to use the db fixture (from quetz.testing.fixtures)"""

    session = get_session(database_url)

    yield session

    session.close()


@pytest.fixture
def dao_extra(db_extra):
    return Dao(db_extra)


@pytest.fixture
def user_with_channel(dao, db, use_migrations, sqlite_in_memory):
    channel_data = rest_models.Channel(name="new-test-channel", private=False)

    user = dao.create_user_with_role("new-user")
    user_id = user.id
    channel = dao.create_channel(channel_data, user_id, "owner")
    db.commit()

    yield user_id
    db.delete(channel)
    db.delete(user)
    db.commit()


# disable running tests in transaction and use on disk database
# because we want to connect to the db with two different
# client concurrently
@pytest.mark.parametrize("sqlite_in_memory", [False])
@pytest.mark.parametrize("auto_rollback", [False])
def test_rollback_on_collision(
    dao: Dao, db, dao_extra, user_with_channel, use_migrations
):
    """testing rollback on concurrent writes."""

    new_package = rest_models.Package(name=f"new-package-{uuid.uuid4()}")

    user_id = user_with_channel
    channel_name = "new-test-channel"

    dao.create_package(channel_name, new_package, user_id, "owner")
    with pytest.raises(errors.DBError, match="(IntegrityError)|(UniqueViolation)"):
        dao_extra.create_package(channel_name, new_package, user_id, "owner")

    requested = db.query(Package).filter(Package.name == new_package.name).one_or_none()

    assert requested

    # need to clean up because we didn't run the test in a transaction

    db.delete(requested)
    db.commit()



# Copyright 2020 QuantStack
# Distributed under the terms of the Modified BSD License.
import logging
import re
from contextlib import contextmanager
from typing import Callable

from sqlalchemy import create_engine, event
from sqlalchemy.engine import Engine
from sqlalchemy.engine.url import make_url
from sqlalchemy.exc import ArgumentError
from sqlalchemy.orm import sessionmaker
from sqlalchemy.orm.session import Session
from sqlalchemy.pool import StaticPool

from quetz.config import Config
from quetz.metrics.middleware import DATABASE_CONNECTIONS_USED, DATABASE_POOL_SIZE

engine = None

logger = logging.getLogger("quetz")


def set_metrics(*args):
    checked_in = engine.pool.checkedin()
    checked_out = engine.pool.checkedout()
    pool_size = checked_in + checked_out
    DATABASE_POOL_SIZE.set(pool_size)
    DATABASE_CONNECTIONS_USED.set(checked_out)


def get_engine(db_url, reuse_engine=True, postgres_kwargs=None, **kwargs) -> Engine:
    if db_url.startswith('sqlite'):
        kwargs.setdefault('connect_args', {'check_same_thread': False})

    if db_url.endswith(':memory:'):
        # If we're using an in-memory database, ensure that only one connection
        # is ever created.
        kwargs.setdefault('poolclass', StaticPool)

    global engine

    if not engine or not reuse_engine:
        if db_url.startswith('postgres'):
            engine = create_engine(
                db_url, **(postgres_kwargs if postgres_kwargs else {}), **kwargs
            )
            for event_name in ['connect', 'close', 'checkin', 'checkout']:
                event.listen(engine, event_name, set_metrics)
        else:
            engine = create_engine(db_url, **kwargs)

        def on_connect(dbapi_conn, conn_record):
            logger.debug("connection opened: %s", engine.pool.status())

        def on_close(dbapi_conn, conn_record):
            logger.debug("connection closed: %s", engine.pool.status())

        event.listen(engine, 'connect', on_connect)
        event.listen(engine, 'close', on_close)

    return engine


def get_session_maker(engine) -> Callable[[], Session]:
    return sessionmaker(autocommit=False, autoflush=True, bind=engine)


def get_session(db_url: str, **kwargs) -> Session:
    """Get a database session.

    Important note: this function is mocked during tests!

    """
    return get_session_maker(get_engine(db_url, **kwargs))()


@contextmanager
def get_db_manager():
    config = Config()
    db = get_session(
        db_url=config.sqlalchemy_database_url,
        echo=config.sqlalchemy_echo_sql,
        postgres_kwargs=dict(
            pool_size=config.sqlalchemy_postgres_pool_size,
            max_overflow=config.sqlalchemy_postgres_max_overflow,
        ),
    )

    try:
        yield db
    finally:
        db.close()


def sanitize_db_url(db_url: str) -> str:
    """
    Sanitizes the DB url so it is safe to print.

    If the URL is parseable with sqlalchemy's make_url,
    it is parsed and the password is replaced.
    If not, we try to replace everything between ":" and "@",
    if those characters are present.

    If neither method succeeds, we give up and return the
    full initial URL.
    """

    # Attempt 1: Actual parsing, this is ideal but may fail
    try:
        parsed_url = make_url(db_url)
        if parsed_url.password:
            return db_url.replace(parsed_url.password, "***")
    except ArgumentError:
        pass

    # Attempt 2: Poor man's parsing: Just replacing everything between ":" and "@"
    if ":" in db_url and "@" in db_url:
        return re.sub(":[^:@]*@", ":***@", db_url)

    # Fallback: We don't understand the URL format, so we do nothing
    return db_url



import time
from typing import Tuple

from prometheus_client import Counter, Gauge, Histogram
from starlette.middleware.base import BaseHTTPMiddleware, RequestResponseEndpoint
from starlette.requests import Request
from starlette.responses import Response
from starlette.routing import Match
from starlette.types import ASGIApp

REQUESTS = Counter(
    "quetz_requests_total",
    "Total count of requests by method and path.",
    ["method", "path_template"],
)
RESPONSES = Counter(
    "quetz_responses_total",
    "Total count of responses by method, path and status codes.",
    ["method", "path_template", "status_code"],
)
REQUESTS_PROCESSING_TIME = Histogram(
    "quetz_requests_processing_time_seconds",
    "Histogram of requests processing time by path (in seconds)",
    ["method", "path_template"],
)
EXCEPTIONS = Counter(
    "quetz_exceptions_total",
    "Total count of exceptions raised by path and exception type",
    ["method", "path_template", "exception_type"],
)
REQUESTS_IN_PROGRESS = Gauge(
    "quetz_requests_in_progress",
    "Gauge of requests by method and path currently being processed",
    ["method", "path_template"],
)
DOWNLOAD_COUNT = Counter(
    "download_count",
    "Total count of package downloads",
    ["channel", "platform", "package_name", "version", "package_type"],
)
UPLOAD_COUNT = Counter(
    "upload_count",
    "Total count of package uploads",
    ["channel", "platform", "package_name", "version", "package_type"],
)

DATABASE_POOL_SIZE = Gauge(
    "database_pool_size", "number of opened database connections"
)

DATABASE_CONNECTIONS_USED = Gauge(
    "database_connections_used", "number of used database connections"
)


class PrometheusMiddleware(BaseHTTPMiddleware):
    def __init__(self, app: ASGIApp, filter_unhandled_paths: bool = False) -> None:
        super().__init__(app)
        self.filter_unhandled_paths = filter_unhandled_paths

    async def dispatch(
        self, request: Request, call_next: RequestResponseEndpoint
    ) -> Response:
        method = request.method
        path_template, is_handled_path = self.get_path_template(request)

        if self._is_path_filtered(is_handled_path):
            return await call_next(request)

        REQUESTS_IN_PROGRESS.labels(method=method, path_template=path_template).inc()
        REQUESTS.labels(method=method, path_template=path_template).inc()

        try:
            before_time = time.perf_counter()
            response = await call_next(request)
            after_time = time.perf_counter()
        except Exception as e:
            EXCEPTIONS.labels(
                method=method,
                path_template=path_template,
                exception_type=type(e).__name__,
            ).inc()
            raise e from None
        else:
            REQUESTS_PROCESSING_TIME.labels(
                method=method, path_template=path_template
            ).observe(after_time - before_time)
            RESPONSES.labels(
                method=method,
                path_template=path_template,
                status_code=response.status_code,
            ).inc()
        finally:
            REQUESTS_IN_PROGRESS.labels(
                method=method, path_template=path_template
            ).dec()

        return response

    @staticmethod
    def get_path_template(request: Request) -> Tuple[str, bool]:
        for route in request.app.routes:
            match, child_scope = route.matches(request.scope)
            if match == Match.FULL:
                return route.path, True

        return request.url.path, False

    def _is_path_filtered(self, is_handled_path: bool) -> bool:
        return self.filter_unhandled_paths and not is_handled_path



# Copyright 2020 QuantStack
# Distributed under the terms of the Modified BSD License.

import bz2
import distutils
import gzip
import hashlib
import inspect
import logging
import secrets
import shlex
import string
import sys
import time
import traceback
import uuid
from datetime import datetime, timezone
from functools import wraps
from pathlib import Path
from typing import Any, Callable
from urllib.parse import unquote

from sqlalchemy import String, and_, cast, collate, not_, or_

from .db_models import Channel, Package, PackageVersion, User


def check_package_membership(package_name, includelist, excludelist):
    if includelist:
        for each_package in includelist:
            if package_name.startswith(each_package):
                return True
        return False
    elif excludelist:
        for each_package in excludelist:
            if package_name.startswith(each_package):
                return False
        return True
    return True


def add_static_file(contents, channel_name, subdir, fname, pkgstore, file_index=None):
    if type(contents) is not bytes:
        raw_file = contents.encode("utf-8")
    else:
        raw_file = contents
    bz2_file = bz2.compress(raw_file)
    gzp_file = gzip.compress(raw_file)

    path = f"{subdir}/{fname}" if subdir else fname
    pkgstore.add_file(bz2_file, channel_name, f"{path}.bz2")
    pkgstore.add_file(gzp_file, channel_name, f"{path}.gz")
    pkgstore.add_file(raw_file, channel_name, f"{path}")

    if file_index:
        add_entry_for_index(file_index, subdir, fname, raw_file)
        add_entry_for_index(file_index, subdir, f"{fname}.bz2", bz2_file)
        add_entry_for_index(file_index, subdir, f"{fname}.gz", gzp_file)


def add_temp_static_file(
    contents, channel_name, subdir, fname, temp_dir, file_index=None
):
    if type(contents) is not bytes:
        raw_file = contents.encode("utf-8")
    else:
        raw_file = contents

    temp_dir = Path(temp_dir)

    if subdir:
        path = temp_dir / channel_name / subdir
    else:
        path = temp_dir / channel_name

    if not path.exists():
        path.mkdir(exist_ok=True, parents=True)

    file_path = path / fname

    with open(file_path, 'wb') as fo:
        fo.write(raw_file)

    bz2_file = bz2.compress(raw_file)
    gzp_file = gzip.compress(raw_file)

    with open(f"{file_path}.bz2", 'wb') as fo:
        fo.write(bz2_file)

    with open(f"{file_path}.gz", 'wb') as fo:
        fo.write(gzp_file)

    if file_index:
        add_entry_for_index(file_index, subdir, fname, raw_file)
        add_entry_for_index(file_index, subdir, f"{fname}.bz2", bz2_file)
        add_entry_for_index(file_index, subdir, f"{fname}.gz", gzp_file)


def add_entry_for_index(files, subdir, fname, data_bytes):
    md5 = hashlib.md5()
    sha = hashlib.sha256()

    md5.update(data_bytes)
    sha.update(data_bytes)

    if subdir:
        files[subdir].append(
            {
                "name": fname,
                "size": len(data_bytes),
                "timestamp": datetime.now(timezone.utc),
                "md5": md5.hexdigest(),
                "sha256": sha.hexdigest(),
            }
        )


def parse_query(search_type, query):
    accepted_filters = []
    if search_type == 'package':
        accepted_filters = [
            'channel',
            'description',
            'summary',
            # 'format',
            'platform',
            # 'version',
            # 'uploader',
        ]
    elif search_type == 'channel':
        accepted_filters = ['description', 'private']
    query = unquote(query.strip())

    args = shlex.split(query)
    keywords = []
    filters = []

    for arg in args:
        if ':' in arg:
            key, val = arg.split(':', 1)
            if (
                key.startswith('-') and key[1:] in accepted_filters
            ) or key in accepted_filters:
                filters.append((key, val.split(',')))
        else:
            arg = arg.strip('"').strip("'")
            keywords.append(arg)

    return keywords, filters


def apply_custom_query(search_type, db, keywords, filters):
    keyword_conditions = []
    negation_argument = None
    each_keyword_condition = None
    for i, each_keyword in enumerate(keywords):
        if each_keyword == 'NOT':
            negation_argument = keywords[i + 1]
            if search_type == 'package':
                each_keyword_condition = Package.name.notlike(f'%{negation_argument}%')
            elif search_type == 'channel':
                each_keyword_condition = collate(Channel.name, "und-x-icu").notlike(
                    f'%{negation_argument}%'
                )
            else:
                raise KeyError(search_type)
        else:
            if each_keyword != negation_argument:
                if search_type == 'package':
                    each_keyword_condition = Package.name.ilike(f'%{each_keyword}%')
                elif search_type == 'channel':
                    each_keyword_condition = collate(Channel.name, "und-x-icu").ilike(
                        f'%{each_keyword}%'
                    )
                else:
                    raise KeyError(search_type)
        keyword_conditions.append(each_keyword_condition)

    query = db.filter(and_(True, *keyword_conditions))

    for each_filter in filters:
        key, values = each_filter
        negate = False
        if key.startswith('-'):
            key = key[1:]
            negate = True
        each_filter_conditions = []
        for each_val in values:
            each_val_condition = None
            each_val = each_val.strip('"').strip("'")
            if search_type == 'package':
                if key == 'channel':
                    each_val_condition = collate(Channel.name, "und-x-icu").ilike(
                        f'%{(each_val)}%'
                    )
                elif key == 'description':
                    each_val_condition = Package.description.contains(each_val)
                elif key == 'summary':
                    each_val_condition = Package.summary.contains(each_val)
                elif key == 'format':
                    each_val_condition = cast(
                        PackageVersion.package_format, String
                    ).ilike(f'%{(each_val)}%')
                elif key == 'platform':
                    each_val_condition = Package.platforms.ilike(f'%{(each_val)}%')
                elif key == 'version':
                    each_val_condition = PackageVersion.version.ilike(f'%{(each_val)}%')
                elif key == 'uploader':
                    each_val_condition = User.username.ilike(f'%{(each_val)}%')
                else:
                    raise KeyError(key)
            elif search_type == 'channel':
                if key == 'description':
                    each_val_condition = Channel.description.contains(each_val)
                elif key == 'private':
                    each_val_condition = Channel.private.is_(
                        bool(distutils.util.strtobool(each_val))
                    )
                else:
                    raise KeyError(key)
            else:
                raise KeyError(search_type)
            each_filter_conditions.append(each_val_condition)
        if negate:
            query = query.filter(not_(or_(*each_filter_conditions)))
        else:
            query = query.filter(or_(*each_filter_conditions))
    return query


class TicToc:
    def __init__(self, description):
        self.description = description

    def __enter__(self):
        self.start = time.time()

    def __exit__(self, ty, val, tb):
        self.stop = time.time()
        print(f"[TOC] {self.description}: {self.stop - self.start}")

    @property
    def elapsed(self):
        return self.stop - self.start


def generate_random_key(length=32):
    alphabet = string.ascii_letters + string.digits
    return ''.join(secrets.choice(alphabet) for i in range(length))


def background_task_wrapper(func: Callable, logger: logging.Logger) -> Callable:
    task_name = func.__name__

    @wraps(func)
    async def wrapper(*args: Any, **kwargs: Any) -> None:
        task_id = uuid.uuid4()

        try:
            func(*args, **kwargs)
            logger.info(f"[{task_id}] Finished {task_name} successfully")
        except Exception as e:
            logger.error(f"[{task_id}] Failed Permanently {task_name} with error: {e}")

            func_args = inspect.signature(func).bind(*args, **kwargs).arguments

            func_args_str = ", ".join(
                f"{name}={repr(value)}" for name, value in func_args.items()
            )

            logger.error(
                f"[{task_id}] Started {task_name} with arguments: {func_args_str}"
            )

            exc_type, exc_value, exc_traceback = sys.exc_info()

            traceback.print_exception(
                exc_type, exc_value, exc_traceback, limit=25, file=sys.stdout
            )

    return wrapper



# Copyright 2020 QuantStack
# Distributed under the terms of the Modified BSD License.
import contextlib
import logging
import os
import random
import shutil
import subprocess
import uuid
from distutils.spawn import find_executable
from enum import Enum
from multiprocessing import get_context
from pathlib import Path
from typing import NoReturn, Optional, Union

import typer
import uvicorn
from alembic import command
from alembic.config import Config as AlembicConfig
from importlib_metadata import entry_points
from sqlalchemy.orm.session import Session
from sqlalchemy_utils.functions import database_exists
from typing_extensions import Annotated

from quetz import pkgstores
from quetz.config import (
    Config,
    _env_config_file,
    _env_prefix,
    configure_logger,
    create_config,
)
from quetz.database import get_session, sanitize_db_url
from quetz.db_models import (
    ApiKey,
    Channel,
    ChannelMember,
    Identity,
    Package,
    PackageMember,
    Profile,
    User,
)

app = typer.Typer()

logger = logging.getLogger("quetz-cli")
configure_logger(loggers=("quetz-cli", "alembic"))


class LogLevel(str, Enum):
    critical = "critical"
    error = "error"
    warning = "warning"
    info = "info"
    debug = "debug"
    trace = "trace"


@contextlib.contextmanager
def working_directory(path):
    """Change working directory and return to previous on exit."""
    prev_cwd = Path.cwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(prev_cwd)


def _alembic_config(db_url: str) -> AlembicConfig:
    script_location = "quetz:migrations"

    migration_modules = [
        f"{ep.module}:versions"
        for ep in entry_points().select(group='quetz.migrations')
    ]
    migration_modules.append("quetz:migrations/versions")

    version_locations = " ".join(migration_modules)

    alembic_cfg = AlembicConfig()
    alembic_cfg.set_main_option('script_location', script_location)
    alembic_cfg.set_main_option('version_locations', version_locations)
    alembic_cfg.set_main_option('sqlalchemy.url', db_url)
    return alembic_cfg


def _run_migrations(
    db_url: Optional[str] = None,
    alembic_config: Optional[AlembicConfig] = None,
    branch_name: str = "heads",
) -> None:
    if db_url:
        if db_url.startswith("postgre"):
            db_engine = "PostgreSQL"
        elif db_url.startswith("sqlite"):
            db_engine = "SQLite"
        else:
            db_engine = db_url.split("/")[0]
        logger.info('Running DB migrations on %s', db_engine)
        if not alembic_config:
            alembic_config = _alembic_config(db_url)
    command.upgrade(alembic_config, branch_name)


def _make_migrations(
    db_url: Optional[str],
    message: str,
    plugin_name: str = "quetz",
    initialize: bool = False,
    alembic_config: Optional[AlembicConfig] = None,
) -> None:
    if not (db_url or alembic_config):
        raise ValueError("provide either alembic_config or db_url")

    found = False

    # import the extra models here to register them with sqlalchemy mapper
    # so that it can create the tables
    from quetz.jobs.models import Job, Task  # noqa
    from quetz.metrics.db_models import PackageVersionMetric  # noqa

    for entry_point in entry_points().select(group='quetz.models'):
        logger.debug("loading plugin %r", entry_point.name)
        entry_point.load()
        if entry_point.name == plugin_name:
            found = True

    if plugin_name != "quetz" and not found:
        raise Exception(
            f"models entrypoint (quetz.models) for plugin {plugin_name} not registered"
        )

    logger.info('Making DB migrations on %r for %r', db_url, plugin_name)
    if not alembic_config and db_url:
        alembic_config = _alembic_config(db_url)

    # find path
    if plugin_name == "quetz":
        version_path = None  # Path(quetz.__file__).parent / 'migrations' / 'versions'
    else:
        entry_point = tuple(
            entry_points().select(group='quetz.migrations', name=plugin_name)
        )[0]
        module = entry_point.load()
        version_path = str(Path(module.__file__).parent / "versions")

    if initialize:
        command.revision(
            alembic_config,
            head="base",
            depends_on="quetz",
            message=message,
            autogenerate=True,
            version_path=version_path,
            branch_label=plugin_name,
            splice=True,
        )
    else:
        command.revision(
            alembic_config,
            head=f"{plugin_name}@head",
            message=message,
            autogenerate=True,
            version_path=version_path,
        )


def _set_user_roles(db: Session, config: Config):
    """Initialize the database and add users from config."""
    if config.configured_section("users"):
        role_map = [
            (config.users_admins, "owner"),
            (config.users_maintainers, "maintainer"),
            (config.users_members, "member"),
        ]

        default_role = config.users_default_role

        for users, role in role_map:
            for username in users:
                try:
                    provider, username = username.split(":")
                except ValueError:
                    # use github as default provider
                    raise ValueError(
                        "could not parse the users setting, please provide users in"
                        "the format 'PROVIDER:USERNAME' where PROVIDER is one of"
                        "'google', 'github', 'dummy', etc."
                    )
                logger.info(f"create user {username} with role {role}")
                user = (
                    db.query(User)
                    .join(Identity)
                    .filter(Identity.provider == provider)
                    .filter(User.username == username)
                    .one_or_none()
                )
                if not user:
                    logger.warning(
                        f"could not find user '{username}' "
                        f"with identity from provider '{provider}'"
                    )
                elif user.role is not None and user.role != default_role:
                    logger.warning(
                        f"user has already role {user.role} not assigning a new role"
                    )
                else:
                    user.role = role

        db.commit()


def _fill_test_database(db: Session) -> None:
    """Create dummy users and channels to allow further testing in dev mode."""
    from quetz.dao import Dao

    test_users = []
    dao = Dao(db)
    try:
        for index, username in enumerate(['alice', 'bob', 'carol', 'dave']):
            user = dao.create_user_with_role(username)

            identity = Identity(
                provider='dummy',
                identity_id=str(index),
                username=username,
            )

            profile = Profile(name=username.capitalize(), avatar_url='/avatar.jpg')

            user.identities.append(identity)  # type: ignore
            user.profile = profile
            db.add(user)
            test_users.append(user)

        for channel_index in range(3):
            channel = Channel(
                name=f'channel{channel_index}',
                description=f'Description of channel{channel_index}',
                private=False,
            )

            for package_index in range(random.randint(30, 60)):
                package = Package(
                    name=f'package{package_index}',
                    summary=f'package {package_index} summary text',
                    description=f'Description of package{package_index}',
                )
                channel.packages.append(package)  # type: ignore

                test_user = test_users[random.randint(0, len(test_users) - 1)]
                package_member = PackageMember(
                    package=package, channel=channel, user=test_user, role='owner'
                )

                db.add(package_member)

            test_user = test_users[random.randint(0, len(test_users) - 1)]

            if channel_index == 0:
                package = Package(name='xtensor', description='Description of xtensor')
                channel.packages.append(package)  # type: ignore

                package_member = PackageMember(
                    package=package, channel=channel, user=test_user, role='owner'
                )

                db.add(package_member)

                # create API key
                key = uuid.uuid4().hex

                key_user = User(id=uuid.uuid4().bytes)
                api_key = ApiKey(
                    key=key, description='test API key', user=test_user, owner=test_user
                )
                db.add(api_key)
                print(f'Test API key created for user "{test_user.username}": {key}')

                key_package_member = PackageMember(
                    user=key_user,
                    channel_name=channel.name,
                    package_name=package.name,
                    role='maintainer',
                )
                db.add(key_package_member)

            db.add(channel)

            channel_member = ChannelMember(
                channel=channel,
                user=test_user,
                role='owner',
            )

            db.add(channel_member)
        db.commit()
    finally:
        db.close()


def _is_deployment(base_dir: Path):
    config_file = base_dir.joinpath("config.toml")
    if base_dir.exists() and base_dir.joinpath("channels").exists():
        # If the config file exists, we assume that the database path
        # is set there (it only matters for sqlite database).
        if config_file.exists():
            config = Config(str(config_file.resolve()))
            db_path = base_dir
        else:
            config = Config()
            db_path = Path(os.getcwd())
        with working_directory(db_path):
            if not database_exists(config.sqlalchemy_database_url):
                logger.error(
                    "Cannot verify that specified database exists. "
                    + sanitize_db_url(config.sqlalchemy_database_url)
                )
                return False
            else:
                return True

    return False


@app.command()
def init_db(
    path: Annotated[str, typer.Argument(help="The path of the deployment")],
):
    """init database and fill users from config file [users] sections"""

    logger.info("Initializing database")

    config = _get_config(path)

    with working_directory(path):
        _run_migrations(config.sqlalchemy_database_url)


@app.command()
def add_user_roles(
    path: Annotated[str, typer.Argument(help="The path of the deployment")],
):
    """Set user roles according to the values from config file [users] sections.

    This command will assign roles only if they were not set before
    (for example using API or an earlier setting in config). The only exception
    is that users who already have a role defined as default_role will have a
    new role set from the config.

    This command will NOT remove roles of existing roles, even if they were
    removed from the config file.
    """

    logger.info("setting up user roles")

    config = _get_config(path)

    with working_directory(path):
        db = get_session(config.sqlalchemy_database_url)
        _set_user_roles(db, config)


@app.command()
def make_migrations(
    path: Annotated[str, typer.Argument(help="The path of the deployment")],
    message: Annotated[str, typer.Option(help="revision message")],
    plugin: Annotated[str, typer.Option(help="head or heads or plugin name")] = "quetz",
    initialize: Annotated[bool, typer.Option(help="initialize migrations")] = False,
):
    """make database migrations for quetz or a plugin"""

    logger.info("Initializing database")

    config = _get_config(path)

    with working_directory(path):
        _make_migrations(config.sqlalchemy_database_url, message, plugin, initialize)


@app.command()
def create(
    path: Annotated[
        str,
        typer.Argument(
            help=(
                "The directory in which the deployment will be created "
                "(will be created if does not exist)"
            ),
        ),
    ],
    copy_conf: Annotated[
        Optional[str],
        typer.Option(help="The configuration to copy from (e.g. dev_config.toml)"),
    ] = None,
    create_conf: Annotated[
        bool,
        typer.Option(
            help="Enable/disable creation of a default configuration file",
        ),
    ] = False,
    delete: Annotated[
        bool,
        typer.Option(
            help="Delete the the deployment if it exists. "
            "Must be specified with --copy-conf or --create-conf",
        ),
    ] = False,
    exists_ok: Annotated[
        bool,
        typer.Option(help="Skip the creation if deployment already exists."),
    ] = False,
    dev: Annotated[
        bool,
        typer.Option(
            help=(
                "Enable/disable dev mode "
                "(fills the database with test data and allows http access)"
            ),
        ),
    ] = False,
):
    """Create a new Quetz deployment."""

    logger.info(f"creating new deployment in path {path}")
    deployment_folder = Path(path).resolve()
    config_file = deployment_folder / "config.toml"

    # Working directory when creating a local database (eg: sqlite).
    # If a config file is created, it will be in the deployment folder,
    # otherwise it will be from cwd.
    db_path = path

    if _is_deployment(deployment_folder):
        if exists_ok:
            logger.info(
                f'Quetz deployment already exists at {deployment_folder}.\n'
                f'Skipping creation.'
            )
            return
        if delete and (copy_conf or create_conf):
            shutil.rmtree(deployment_folder)
        else:
            typer.echo(
                'Use the start command to start a deployment '
                'or specify --delete with --copy-conf or --create-conf.',
                err=True,
            )
            raise typer.Abort()

    deployment_folder.mkdir(parents=True, exist_ok=True)

    # only authorize path with a config file to avoid deletion of unexpected files
    # when deleting Quetz instance
    if any(f != config_file for f in deployment_folder.iterdir()):
        typer.echo(
            f'Quetz deployment not allowed at {path}.\n'
            'The path should not contain more than the configuration file.',
            err=True,
        )
        raise typer.Abort()

    if not config_file.exists() and not create_conf and not copy_conf:
        try:
            # If no config file is provided or created, try to get config
            # from environmental variables.
            # If the config file does not exist, we assume that the database URL
            # is set from current directory (it only matters for sqlite database).
            config = Config()
            db_path = os.getcwd()
        except Exception as e:
            logger.warning(msg=e)
            typer.echo(
                'No configuration file provided.\n'
                'Use --create-conf or --copy-conf to produce a config file\n'
                'or set up $QUETZ_SQLALCHEMY_DATABASE_URL and $QUETZ_SESSION_SECRET',
                err=True,
            )
            raise typer.Abort()

    if copy_conf:
        if not os.path.exists(copy_conf):
            typer.echo(f'Config file to copy does not exist {copy_conf}.', err=True)
            raise typer.Abort()

        typer.echo(f"Copying config file from {copy_conf} to {config_file}")
        shutil.copyfile(copy_conf, config_file)

    if not config_file.exists() and create_conf:
        https = 'false' if dev else 'true'
        conf = create_config(https=https)
        with open(config_file, 'w') as f:
            f.write(conf)

    os.environ[_env_prefix + _env_config_file] = str(config_file.resolve())
    config = Config(str(config_file))

    deployment_folder.joinpath('channels').mkdir(exist_ok=True)

    with working_directory(db_path):
        db = get_session(config.sqlalchemy_database_url)
        _run_migrations(config.sqlalchemy_database_url)
        if dev:
            _fill_test_database(db)
        _set_user_roles(db, config)


def _get_config(path: Union[Path, str]) -> Config:
    """get config path"""
    config_file = Path(path) / 'config.toml'
    config = Config(str(config_file.resolve()))
    if not os.environ.get(_env_prefix + _env_config_file):
        os.environ[_env_prefix + _env_config_file] = str(config_file.resolve())

    return config


@app.command()
def start(
    path: Annotated[
        Optional[str], typer.Argument(help="The path of the deployment")
    ] = None,
    port: Annotated[int, typer.Option(help="The port to bind")] = 8000,
    host: Annotated[
        str, typer.Option(help="The network interface to bind")
    ] = "127.0.0.1",
    proxy_headers: Annotated[
        bool, typer.Option(help="Enable/disable X-Forwarded headers")
    ] = True,
    log_level: Annotated[
        LogLevel,
        typer.Option(
            help="Set the logging level",
        ),
    ] = LogLevel.info,
    reload: Annotated[
        bool,
        typer.Option(
            help=(
                "Enable/disable automatic reloading of the server when sources are "
                "modified"
            ),
        ),
    ] = False,
    supervisor: Annotated[
        bool,
        typer.Option(
            help="Start job supervisor with quetz",
        ),
    ] = True,
) -> NoReturn:
    """Start a Quetz deployment.

    To be started, a deployment has to be already created.
    At this time, only Uvicorn is supported as manager.
    """

    logger.info(f"deploying quetz from directory {path}")

    if path:
        deployment_folder = Path(path)
    else:
        deployment_folder = Path("")
        path = os.getcwd()

    config = _get_config(deployment_folder)

    if not _is_deployment(deployment_folder):
        if isinstance(config.get_package_store(), pkgstores.LocalStore):
            typer.echo(
                'The specified directory is not a deployment and the package store '
                'is set as local.\n'
                'Use the create or run command to create a deployment.',
                err=True,
            )
            raise typer.Abort()

    supervisor_process = None
    if supervisor:
        logger.info("starting supervisor")
        ctx = get_context("spawn")
        supervisor_process = ctx.Process(
            target=start_supervisor_daemon,
            args=(deployment_folder,),
        )
        supervisor_process.start()

    with working_directory(path):
        import quetz

        quetz_src = os.path.dirname(quetz.__file__)
        uvicorn.run(
            "quetz.main:app",
            reload=reload,
            reload_dirs=(quetz_src,),
            port=port,
            proxy_headers=proxy_headers,
            host=host,
            log_level=log_level,
        )

    if supervisor_process is not None:
        supervisor_process.terminate()
        supervisor_process.join()


@app.command()
def run(
    path: Annotated[str, typer.Argument(help="The path of the deployment")],
    copy_conf: Annotated[
        Optional[str],
        typer.Option(help="The configuration to copy from (e.g. dev_config.toml)"),
    ] = None,
    create_conf: Annotated[
        bool,
        typer.Option(
            help="Enable/disable creation of a default configuration file",
        ),
    ] = False,
    delete: Annotated[
        bool,
        typer.Option(
            help="Delete the the deployment if it exists. "
            "Must be specified with --copy-conf or --create-conf",
        ),
    ] = False,
    skip_if_exists: Annotated[
        bool,
        typer.Option(help="Skip the creation if deployment already exists."),
    ] = False,
    dev: Annotated[
        bool,
        typer.Option(
            help=(
                "Enable/disable dev mode "
                "(fills the database with test data and allows http access)"
            ),
        ),
    ] = False,
    port: Annotated[int, typer.Option(help="The port to bind")] = 8000,
    host: Annotated[
        str, typer.Option(help="The network interface to bind")
    ] = "127.0.0.1",
    proxy_headers: Annotated[
        bool, typer.Option(help="Enable/disable X-Forwarded headers")
    ] = True,
    log_level: Annotated[
        LogLevel,
        typer.Option(
            help="Set the logging level",
        ),
    ] = LogLevel.info,
    reload: Annotated[
        bool,
        typer.Option(
            help="Enable/disable automatic reloading of the server when sources are "
            "modified",
        ),
    ] = False,
) -> NoReturn:
    """Run a Quetz deployment.

    It performs sequentially create and start operations."""

    abs_path = os.path.abspath(path)
    create(abs_path, copy_conf, create_conf, delete, skip_if_exists, dev)
    start(abs_path, port, host, proxy_headers, log_level, reload)


@app.command()
def delete(
    path: Annotated[str, typer.Argument(help="The path of the deployment")],
    force: Annotated[
        bool,
        typer.Option(help="Enable/disable removal without confirmation prompt"),
    ] = False,
) -> None:
    """Delete a Quetz deployment."""

    deployment_dir = Path(path)
    if not _is_deployment(deployment_dir):
        typer.echo(f'No Quetz deployment found at {path}.', err=True)
        raise typer.Abort()

    if not force and not typer.confirm(f"Delete Quetz deployment at {path}?"):
        raise typer.Abort()

    shutil.rmtree(deployment_dir)


@app.command()
def plugin(
    cmd: str,
    path: Annotated[str, typer.Argument(help="Path to the plugin folder")],
) -> None:
    if cmd == 'install':
        abs_path = Path(path).absolute()
        if not (abs_path / "setup.py").exists():
            raise ValueError(f"Could not find any setup.py file in {abs_path}/.")
        requirements = abs_path / "conda-requirements.txt"

        conda_exes = ['mamba', 'conda', 'micromamba']
        conda_exe_path = None
        if requirements.exists():
            for conda_exe in conda_exes:
                conda_exe_path = find_executable(conda_exe)
                if conda_exe_path:
                    break

            if conda_exe_path is None:
                print(
                    f"Could not find any of {', '.join(conda_exes)}. "
                    f"One is needed to install the plugin requirements."
                )
                exit(1)

            print(f"Installing requirements.txt for {os.path.split(abs_path)[1]}")
            subprocess.call(
                [
                    conda_exe_path,
                    'install',
                    '--channel',
                    'conda-forge',
                    '--file',
                    requirements,
                ]
            )

        pip_exe_path = find_executable('pip')
        if pip_exe_path is None:
            # Try to install pip if it's missing
            if conda_exe_path is not None:
                print("pip is missing, installing...")
                subprocess.call(
                    [conda_exe_path, 'install', '--channel', 'conda-forge', 'pip']
                )
                pip_exe_path = find_executable('pip')

        if pip_exe_path is None:
            print("Could not find pip to install the plugin.")
            exit(1)
        subprocess.call([pip_exe_path, 'install', abs_path])
    else:
        print(f"Command '{cmd}' not yet understood.")


def start_supervisor_daemon(path: Path, num_procs=None):
    from quetz.jobs.runner import Supervisor
    from quetz.tasks.workers import get_worker

    configure_logger(loggers=("quetz",))
    config = _get_config(path)
    manager = get_worker(config, num_procs=num_procs)

    # If the config file exists, we assume that the database path
    # is set there (it only matters for sqlite database).
    db_path = path if path.joinpath("config.toml").exists() else os.getcwd()
    with working_directory(db_path):
        db = get_session(config.sqlalchemy_database_url)
        supervisor = Supervisor(db, manager)
        try:
            supervisor.run()
        except KeyboardInterrupt:
            logger.info("stopping supervisor")
        finally:
            db.close()


@app.command()
def watch_job_queue(
    path: Annotated[str, typer.Argument(help="Path to the plugin folder")],
    num_procs: Annotated[
        Optional[int],
        typer.Option(help="Number of processes to use. Default: number of CPU cores"),
    ] = None,
) -> None:
    start_supervisor_daemon(Path(path), num_procs)


if __name__ == "__main__":
    app()



from quetz.metrics import tasks as metrics_tasks
from quetz.tasks import cleanup, indexing, mirror, reindexing

JOB_HANDLERS = {
    "synchronize": mirror.synchronize_packages,
    "synchronize_repodata": mirror.synchronize_packages,
    "validate_packages": indexing.validate_packages,
    "generate_indexes": indexing.update_indexes,
    "reindex": reindexing.reindex_packages_from_store,
    "synchronize_metrics": metrics_tasks.synchronize_metrics_from_mirrors,
    "pkgstore_cleanup": cleanup.cleanup_channel_db,
    "db_cleanup": cleanup.cleanup_temp_files,
    "pkgstore_cleanup_dry_run": cleanup.cleanup_channel_db,
    "db_cleanup_dry_run": cleanup.cleanup_temp_files,
}
